import io
import subprocess
import sys
from pathlib import Path

import pytest

from argparse_to_md.formatter import MarkdownHelpFormatterOptions
from argparse_to_md.loader import FunctionLoader
from argparse_to_md.markdown_processor import args_to_options, process_markdown


def test_usage():
    data_dir = Path(__file__).parent / "data"
    out_md = io.StringIO()
    loader = FunctionLoader()
    with open(data_dir / "test1.md.in") as in_md:
        process_markdown(in_md, out_md, loader)

    with open(data_dir / "test1.md.actual", "w") as actual_md:
        actual_md.write(out_md.getvalue())

    assert out_md.getvalue() == (data_dir / "test1.md.expected").read_text()


def test_cli_check_uage():
    test_dir = Path(__file__).parent

    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "argparse_to_md",
            "--check",
            "-i",
            str(test_dir / "data" / "test2.md.in"),
        ],
        text=True,
        capture_output=True,
    )
    output = result.stderr

    assert result.returncode == 2
    assert "Changes required in" in output
    assert "Updating" not in output
    assert "-- `--foo FOO`: foo no help" in output
    assert "+- `--foo FOO`: foo help" in output

    result = subprocess.run(
        [
            sys.executable,
            "-m",
            "argparse_to_md",
            "--check",
            "-i",
            str(test_dir / "data" / "test2.md.expected"),
        ],
        text=True,
        capture_output=True,
    )

    assert result.returncode == 0
    assert result.stderr == ""
    assert result.stdout == ""


def test_subparsers():
    data_dir = Path(__file__).parent / "data"
    out_md = io.StringIO()
    loader = FunctionLoader()
    with open(data_dir / "test3.md.in") as in_md:
        process_markdown(in_md, out_md, loader)

    with open(data_dir / "test3.md.actual", "w") as actual_md:
        actual_md.write(out_md.getvalue())

    assert out_md.getvalue() == (data_dir / "test3.md.expected").read_text()


def test_arguments_to_options():
    assert args_to_options("") == MarkdownHelpFormatterOptions()
    assert args_to_options("subheading_level=2") == MarkdownHelpFormatterOptions(subheading_level=2)
    with pytest.raises(ValueError):
        args_to_options("subheading_level=2:foo=bar")
    with pytest.raises(ValueError):
        args_to_options("opt1:opt2")
    with pytest.raises(ValueError):
        args_to_options("opt=val=val2")



import argparse

import module_which_does_not_exist


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog='testprog', description='Description of the program')
    parser.add_argument('--foo', help='foo help')
    parser.add_argument('--bar', help='bar help')
    parser.add_argument('--baz', help='baz help')
    parser.add_argument('--longer-argument', help='longer-argument help')
    parser.add_argument('--even-longer-argument', help='even-longer-argument help')

    return parser



import os
import re
import typing as t

from .formatter import MarkdownHelpFormatterOptions, gen_argparse_help
from .loader import FunctionLoader


def process_markdown(in_markdown: t.TextIO, out_markdown: t.TextIO, loader: FunctionLoader) -> None:
    """
    Process the input markdown file, updating the argparse help text in the file.

    :param in_markdown: Input markdown file
    :param out_markdown: Output markdown file
    :param loader: FunctionLoader instance to load the argparse factory function
    """
    # Read the input file, processing each line:
    # - if we are not processing a block of argparse help text, just copy the line to the output
    # - if we encounter and argparse_doc comment, start generating argparse help text
    # - skip all lines until we encounter argparse_doc_end comment

    in_argparse_to_md_block = False

    # Match comments like <!--argparse_to_md:test3:get_parser:arg1=val1:arg2=val2-->
    argparse_doc_regex = re.compile(r"<!--\s*argparse_to_md:(?P<module>[\w.]+):(?P<function>\w+)(?P<args>:.*)?\s*-->")
    argparse_doc_end_regex = re.compile(r"<!--\s*argparse_to_md_end\s*-->")

    # Get the current working directory of the input file, so we can add it to the sys.path
    cwd = None
    if hasattr(in_markdown, "name"):
        cwd = os.path.dirname(in_markdown.name)

    for line in in_markdown.readlines():
        if not in_argparse_to_md_block:
            out_markdown.write(line)
            match = argparse_doc_regex.match(line)
            if match:
                in_argparse_to_md_block = True
                module = match.group("module")
                function = match.group("function")
                args = match.group("args")
                parser_factory_function = loader.load_function(module, function, cwd)
                parser = parser_factory_function()
                gen_argparse_help(parser, out_markdown, args_to_options(args))
        else:
            match = argparse_doc_end_regex.match(line)
            if match:
                in_argparse_to_md_block = False
                out_markdown.write(line)
            else:
                continue


def args_to_options(args: str) -> MarkdownHelpFormatterOptions:
    if not args:
        return MarkdownHelpFormatterOptions()

    args = args.strip(":")
    args_dict = {}
    for arg in args.split(":"):
        if "=" not in arg or arg.count("=") != 1:
            raise ValueError(f"Invalid argument format: '{arg}'. Expected format is key=value.")
        key, value = arg.split("=", 1)
        args_dict[key] = value

    subheading_level = 0
    if "subheading_level" in args_dict:
        subheading_level = int(args_dict["subheading_level"])
        del args_dict["subheading_level"]

    if args_dict:
        raise ValueError(f"Unknown arguments: {args_dict}")

    return MarkdownHelpFormatterOptions(subheading_level=subheading_level)



import argparse
import typing as t
from dataclasses import dataclass

HELP_WIDTH = 100


@dataclass
class MarkdownHelpFormatterOptions:
    subheading_level: int = 0


class MarkdownHelpFormatter(argparse.HelpFormatter):
    def __init__(
        self,
        prog,
        indent_increment=2,
        max_help_position=24,
        width=HELP_WIDTH,
        usage_prefix: str = "Usage:",
    ):
        self.usage_prefix = usage_prefix
        super().__init__(prog=prog, indent_increment=indent_increment, max_help_position=max_help_position, width=width)

    def _format_usage(self, usage, actions, groups, prefix):
        prefix = f"{self.usage_prefix}\n```\n"
        result = super()._format_usage(usage, actions, groups, prefix).rstrip("\n")
        result += "\n```\n"
        return result

    def _format_action(self, action):
        if action.dest == "help":
            return ""
        option_decl = self._format_action_invocation(action)
        has_curly_braces = False
        if option_decl.startswith("{") and option_decl.endswith("}"):
            has_curly_braces = True
            option_decl = option_decl[1:-1]

        option_decl_parts = [od.strip() for od in option_decl.split(",")]
        option_decl_md = ", ".join([f"`{od}`" for od in option_decl_parts])
        if has_curly_braces:
            option_decl_md = f"{{{option_decl_md}}}"
        return f"- {option_decl_md}: {action.help}\n"


def get_formatter_with_usage(usage_prefix: str):
    def formatter(prog, indent_increment=2, max_help_position=24, width=HELP_WIDTH):
        return MarkdownHelpFormatter(prog, indent_increment, max_help_position, width, usage_prefix)

    return formatter


def gen_argparse_help(parser: argparse.ArgumentParser, out_readme: t.TextIO, options: MarkdownHelpFormatterOptions):
    if options.subheading_level > 0:
        subheading_prefix = "#" * options.subheading_level + " "
    else:
        subheading_prefix = ""

    parser.formatter_class = get_formatter_with_usage(f"{subheading_prefix}Usage:")
    parser._optionals.title = "Optional arguments"  # pylint: disable=protected-access
    parser._positionals.title = "Positional arguments"  # pylint: disable=protected-access
    parser.print_help(out_readme)

    subparsers_actions = [action for action in parser._actions if isinstance(action, argparse._SubParsersAction)]
    for subparsers_action in subparsers_actions:
        # get all subparsers and print help
        for choice, subparser in subparsers_action.choices.items():
            out_readme.write("\n")
            subparser.formatter_class = get_formatter_with_usage(f"{subheading_prefix}Usage of `{choice}`:\n")
            subparser._optionals.title = f"Optional arguments of `{choice}`"  # pylint: disable=protected-access
            subparser._positionals.title = f"Positional arguments of `{choice}`"  # pylint: disable=protected-access
            subparser.print_help(out_readme)



import argparse
import difflib
import io
import sys

from . import __version__
from .loader import FunctionLoader
from .markdown_processor import process_markdown


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog="argparse_to_md")
    parser.add_argument(
        "-i",
        "--input",
        nargs="+",
        type=argparse.FileType("r+"),
        help="Markdown file to update (can be specified multiple times).",
    )
    parser.add_argument(
        "--extra-sys-path", nargs="+", help="Extra paths to add to PYTHONPATH before loading the module"
    )
    parser.add_argument(
        "--check",
        action="store_true",
        help="Check if the files need to be updated, but don't modify them. "
        "Non-zero exit code is returned if any file needs to be updated.",
    )
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}")
    return parser


def main() -> None:
    parser = get_parser()
    args = parser.parse_args()

    if not args.input:
        raise SystemExit("No input files specified")

    loader = FunctionLoader(args.extra_sys_path)

    changes_required = False
    for in_markdown in args.input:
        in_markdown_str = in_markdown.read()
        in_markdown.seek(0)
        out_markdown = io.StringIO()
        process_markdown(in_markdown, out_markdown, loader)
        out_markdown_str = out_markdown.getvalue()

        if in_markdown_str != out_markdown_str:
            if args.check:
                print(f"Changes required in {in_markdown.name}:", file=sys.stderr)
                for line in difflib.unified_diff(
                    in_markdown_str.splitlines(), out_markdown_str.splitlines(), lineterm=""
                ):
                    print(line, file=sys.stderr)
                changes_required = True
            else:
                print(f"Updating {in_markdown.name}...", file=sys.stderr)
                in_markdown.seek(0)
                in_markdown.truncate()
                in_markdown.write(out_markdown.getvalue())
                in_markdown.close()

    if args.check and changes_required:
        raise SystemExit(2)


main()



try:
    from ._version import __version__
except ImportError:
    # This happens when the package is not installed, e.g. when running pre-commit in CI
    __version__ = "0.0.0"

from .loader import FunctionLoader
from .markdown_processor import process_markdown

__all__ = ["process_markdown", "FunctionLoader", "__version__"]



import argparse

def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(prog='testprog')
    subparsers = parser.add_subparsers(dest='action', help='action to run')

    foo_action = subparsers.add_parser('foo', description='foo action')
    foo_action.add_argument('--opt1', help='option 1')
    foo_action.add_argument('--opt2', help='option 2')

    bar_action = subparsers.add_parser('bar', description='bar action')
    bar_action.add_argument('--opt3', help='option 3')
    bar_action.add_argument('--opt4', help='option 4')

    parser.add_argument('--opt5', help='option 5')

    return parser

if __name__ == '__main__':
    parser = get_parser()
    parser.parse_args()



import importlib
import sys
import typing as t
from unittest.mock import MagicMock


class FunctionLoader:
    def __init__(self, extra_sys_path: t.Optional[t.List[str]] = None):
        self.extra_sys_path = extra_sys_path or []
        self.modules_imported: t.Dict[str, t.Any] = {}

    @staticmethod
    def _sys_path_extend(extra_sys_path) -> t.ContextManager[None]:
        class SysPathContext:
            def __enter__(self):
                if extra_sys_path:
                    sys.path.extend(extra_sys_path)

            def __exit__(self, exc_type, exc_value, traceback):
                if extra_sys_path:
                    for path in extra_sys_path:
                        sys.path.remove(path)

        return SysPathContext()

    def load_function(self, module_name: str, function_name: str, cwd: t.Optional[str] = None) -> t.Any:
        """
        Load a function from a module by name.

        :param module_name: The name of the module to import.
        :param function_name: The name of the function to load from the module.
        :param cwd: Optional: the current working directory, to add to the sys.path.
        """
        extra_sys_path = self.extra_sys_path
        if cwd is not None:
            extra_sys_path = [cwd] + extra_sys_path
        with self._sys_path_extend(extra_sys_path):
            last_missing_module_name = None
            while True:
                try:
                    if module_name in self.modules_imported:
                        module = self.modules_imported[module_name]
                        break

                    module = importlib.import_module(module_name)
                    self.modules_imported[module_name] = module
                    break

                except ImportError as e:
                    err = str(e)

                    if "No module named" in err:
                        # find the module name in the error message
                        missing_module_name = err.split("'")[1]
                        if missing_module_name == module_name:
                            print(f"Error importing module {module_name}", file=sys.stderr)
                            raise e
                        if missing_module_name == last_missing_module_name:
                            print(f"Error importing module {missing_module_name} after adding a mock", file=sys.stderr)
                            raise e
                        # create a mock module
                        missing_module = MagicMock()
                        # try again to import the module
                        print(f"Note: creating mock module {missing_module_name}", file=sys.stderr)
                        sys.modules[missing_module_name] = missing_module
                        continue

                    else:
                        raise e

            return getattr(module, function_name)
